unit frQueue;

{.$DEFINE NOQUEUESTORE}

interface
uses
     PlugAPI,
     PlugSIDE,
     PlugUTIL,

     Objects,

     RStream,

     Thunks,

     jtMHLcmn,

     frFwdReq,
     frOrder,

     rlopAPI,

     Wizard;

const
 ForwardRequestQueueID          = 'ForwardRequestQueue';

type
 PForwardRequestQueueItem = ^TForwardRequestQueueItem;
 TForwardRequestQueueItem = object(TSortedCollection)
  Name, Desc: PString;
  RequestedBy, RequestedFrom: Pointer;
  RequestedAt: Longint; { julian day }

  constructor Init(const AName, ADesc: String; const ARequestedBy, ARequestedFrom: Pointer);
  constructor Load(var S: TStream);
  procedure Store(var S: TStream); virtual;
  function Compare(Key1, Key2: Pointer): xInteger; virtual;
  procedure FreeItem(Item: Pointer); virtual;
  procedure ResetRequestedAt;
  destructor Done; virtual;
 end;

 PForwardRequestQueue = ^TForwardRequestQueue;
 TForwardRequestQueue = object(TStringCollection)
  procedure Load(var S: TStream); virtual;
  procedure Store(var S: TStream); virtual;
  function KeyOf(Item: Pointer): Pointer; virtual;
  procedure FreeItem(Item: Pointer); virtual;
 public
  procedure AddRequest(const By: Pointer; const Query: PForwardRequestQuery);
  procedure KillRequest(const QueueItem: PForwardRequestQueueItem; const Details: String; const Exclude: Pointer);
  procedure NotifyRequest(const Action: TForwardRequestAction; const QueueItem: PForwardRequestQueueItem;
   const Details: String);
  procedure CheckExpiredRequests;
  function MaxRequestedByCheck(const Link: Pointer): Boolean;
  function MaxRequestedToCheck(const Link: Pointer): Boolean;
 end;

procedure LoadQueue;
procedure StoreQueue;

implementation
uses
     frCmn,
     frAnn;

{ TForwardRequestQueueItem }

constructor TForwardRequestQueueItem.Init(const AName, ADesc: String; const ARequestedBy, ARequestedFrom: Pointer);
 begin
  inherited Init($4, $4);

  Name:=NewStr(AName);
  Desc:=NewStr(ADesc);

  RequestedBy:=ARequestedBy;
  RequestedFrom:=ARequestedFrom;

  Insert(RequestedBy);
  Insert(RequestedFrom);

  ResetRequestedAt;
 end;

constructor TForwardRequestQueueItem.Load(var S: TStream);
 var
  Countdown: Longint;
  RequestedByAddress, RequestedFromAddress, LinkAddress: TAddress;
  Link: Pointer;
 begin
  S.Read(Countdown, SizeOf(Countdown));

  inherited Init(Countdown, $4);

  Name:=S.ReadStr;
  Desc:=S.ReadStr;

  S.Read(RequestedAt, SizeOf(RequestedAt));
  S.Read(RequestedByAddress, SizeOf(RequestedByAddress));
  S.Read(RequestedFromAddress, SizeOf(RequestedFromAddress));

  RequestedBy:=jt.link.Seek(RequestedByAddress);
  RequestedFrom:=jt.link.Seek(RequestedFromAddress);

  while Countdown > 0 do
   begin
    S.Read(LinkAddress, SizeOf(LinkAddress));

    Link:=jt.link.Seek(LinkAddress);

    if Link <> nil then
     Insert(Link)
    else
     if (RequestedBy <> nil) and (RequestedFrom <> nil) then
      jt.LogWrite('?', 'Strange request for ' + Name^ + ': pending subscriber dead (' + AddressToStrEx(LinkAddress) + ')');

    Dec(Countdown);
   end;

  if RequestedBy = nil then
   begin
    jt.LogWrite('?', 'Lost request for ' + Name^ + ': requested-by dead (' + AddressToStrEx(RequestedByAddress) + ')');

    Done;

    Fail;
   end;

  if RequestedFrom = nil then
   begin
    jt.LogWrite('?', 'Lost request for ' + Name^ + ': requested-from dead (' + AddressToStrEx(RequestedFromAddress) + ')');

    Done;

    Fail;
   end;
 end;

procedure TForwardRequestQueueItem.Store(var S: TStream);
 var
  Countup: Longint;
  Address: TAddress;
 procedure Perform(const Link: Pointer); far;
  begin
   jt.link.GetAddress(Link, Address);

   S.Write(Address, SizeOf(Address));
  end;
 begin
  Countup:=Count;

  S.Write(Countup, SizeOf(Countup));

  S.WriteStr(Name);
  S.WriteStr(Desc);

  S.Write(RequestedAt, SizeOf(RequestedAt));

  jt.link.GetAddress(RequestedBy, Address);

  S.Write(Address, SizeOf(Address));

  jt.link.GetAddress(RequestedFrom, Address);

  S.Write(Address, SizeOf(Address));

  ForEach(@Perform);
 end;

function TForwardRequestQueueItem.Compare(Key1, Key2: Pointer): xInteger;
 var
  K1: Longint absolute Key1;
  K2: Longint absolute Key2;
 begin
  Compare:=CompareLongint(K1, K2);
 
{
  if K1 < K2 then Compare:=-1 else
  if K1 > K2 then Compare:=1 else
   Compare:=0;
}
 end;

procedure TForwardRequestQueueItem.FreeItem(Item: Pointer);
 begin
 end;

procedure TForwardRequestQueueItem.ResetRequestedAt;
 var
  DT: TDateTime;
 begin
  GetCurrentDateTime(DT);

  RequestedAt:=GregorianToJulian(DT);
 end;

destructor TForwardRequestQueueItem.Done;
 begin
  DisposeStr(Desc);
  DisposeStr(Name);

  inherited Done;
 end;

{ TForwardRequestQueue }

procedure TForwardRequestQueue.Load(var S: TStream);
 var
  Countdown: Longint;
  Item: PForwardRequestQueueItem;
 begin
  FreeAll;

  S.Read(Countdown, SizeOf(Countdown));

  while Countdown > 0 do
   begin
    Item:=New(PForwardRequestQueueItem, Load(S));

    if Item <> nil then
     Insert(Item);

    Dec(Countdown);
   end;
 end;

procedure TForwardRequestQueue.Store(var S: TStream);
 var
  Countup: Longint;
 procedure Perform(const Item: PForwardRequestQueueItem); far;
  begin
   Item^.Store(S);
  end;
 begin
  Countup:=Count;

  S.Write(Countup, SizeOf(Countup));

  ForEach(@Perform);
 end;

function TForwardRequestQueue.KeyOf(Item: Pointer): Pointer;
 begin
  KeyOf:=PForwardRequestQueueItem(Item)^.Name;
 end;

procedure TForwardRequestQueue.FreeItem(Item: Pointer);
 begin
  TCollection.FreeItem(Item);
 end;

{ public methods }

procedure TForwardRequestQueue.AddRequest(const By: Pointer; const Query: PForwardRequestQuery);
 var
  KilledArea: Pointer;
  QueueItem: PForwardRequestQueueItem;
 begin
  KilledArea:=jt.area.killed.Seek(Query^.Item^.Name^);

  if KilledArea <> nil then
   begin
    jt.area.killed.Destroy(KilledArea);

    jt.area.SetChanged(acfChanged, 'addreq/killedarea');
   end;

  QueueItem:=New(PForwardRequestQueueItem, Init(Query^.Item^.Name^,
                                                GetPString(Query^.Item^.Desc),
                                                By,
                                                Query^.fr^.Link));

  Insert(QueueItem);

  rlopSendAreaSubscribe(Query^.fr^.Link, Query^.Item^.Name^);

  NotifyRequest(frRequested, QueueItem, '');
 end;

procedure TForwardRequestQueue.KillRequest(const QueueItem: PForwardRequestQueueItem; const Details: String;
  const Exclude: Pointer);
 procedure Unsubscribe(const Link: Pointer); far;
  var
   rc: String;
  begin
   if rlopIsLinkValid(Link, rc) and (Link <> Exclude) then
    rlopSendAreaUnsubscribe(Link, QueueItem^.Name^);
  end;
 begin
  NotifyRequest(frKilled, QueueItem, Details);

  QueueItem^.ForEach(@Unsubscribe);

  jt.area.killed.Create(QueueItem^.Name^);

  jt.area.SetChanged(acfChanged, 'killreq/killedareacreate');

  Free(QueueItem);
 end;

procedure TForwardRequestQueue.NotifyRequest(const Action: TForwardRequestAction; const QueueItem: PForwardRequestQueueItem;
  const Details: String);
 begin
  AddAnnounce(Action, QueueItem, Details);
 end;

procedure TForwardRequestQueue.CheckExpiredRequests;
 const
  First: Boolean = True;
 var
  ThisDay: Longint;
 procedure CheckFirst;
  begin
   if First then
    begin
     jt.LogWrite('F', 'Checking expired forward requests');

     First:=False;
    end;
  end;
 function CheckExpired(const Item: PForwardRequestQueueItem): Boolean; far;
  var
   fr: PForwardRequests;
   rc: String;
   Address: TAddress;
   Queries: PForwardRequestQueries;
   Query: PForwardRequestQuery;
   Index: xInteger;
  function LookupQuery(const Query: PForwardRequestQuery): Boolean; far;
   begin
    LookupQuery:=Item^.IndexOf(Query^.fr^.Link) = -1;
   end;
  begin
   CheckExpired:=True;

   fr:=GetForwardRequests(Item^.RequestedFrom);

   if not fr^.Enabled then
    begin
     CheckFirst;

     jt.LogWrite('F', Item^.Name^ + ': invalid request (requested-from link does not support fwdreq)');

     KillRequest(Item, 'invalid req/fr', nil);
    end else
   if not rlopIsLinkValid(Item^.RequestedFrom, rc) then
    begin
     CheckFirst;

     jt.LogWrite('F', Item^.Name^ + ': invalid request (requested-from link does not support RLOP: ' + rc + ')');

     KillRequest(Item, 'invalid req/rlop', nil);
    end
   else
    if ThisDay - Item^.RequestedAt >= fr^.GetTimeout then
     begin
      Queries:=New(PForwardRequestQueries, Init($10, $10));

      fwdreqOrder^.PerformQuery(Item^.Name^, Queries, nil);

      Query:=Queries^.Seek(fr^.Address);

      if Query <> nil then
       for Index:=0 to Queries^.IndexOf(Query) do
        Queries^.AtFree(0);

      Query:=Queries^.FirstThat(@LookupQuery);

      if Query <> nil then
       begin
        jt.link.GetAddress(Item^.RequestedFrom, Address);

        rlopSendAreaUnsubscribe(Item^.RequestedFrom, Item^.Name^);

        Item^.Free(Item^.RequestedFrom);

        Item^.RequestedFrom:=Query^.fr^.Link;

        Item^.Insert(Item^.RequestedFrom);

        rlopSendAreaSubscribe(Item^.RequestedFrom, Item^.Name^);

        NotifyRequest(frRerequested, Item, 'timeout ' + AddressToStrEx(Address));

        CheckFirst;

        jt.LogWrite('F', Item^.Name^ + ': re-requested from ' + AddressToStrEx(Query^.fr^.Address));

        Item^.ResetRequestedAt;

        CheckExpired:=False;
       end
      else
       begin
        CheckFirst;

        jt.LogWrite('F', Item^.Name^ + ': forward request expired, no links to re-request');

        KillRequest(Item, 'timeout', nil);
       end;

      Dispose(Queries, Done);
     end
    else
     CheckExpired:=False;
  end;
 var
  Item: PForwardRequestQueueItem;
  DT: TDateTime;
  K: Longint;
 begin
  GetCurrentDateTime(DT);

  ThisDay:=GregorianToJulian(DT);

  K:=0;

  repeat
   Item:=At(K);

   if Item <> nil then
    if CheckExpired(Item) then
     Continue;

   Inc(K);
  until Item = nil;
 end;

function TForwardRequestQueue.MaxRequestedByCheck(const Link: Pointer): Boolean;
 var
  Counter: Longint;
 procedure Perform(const Item: PForwardRequestQueueItem); far;
  begin
   if Item^.RequestedBy = Link then
    Inc(Counter);
  end;
 var
  Address: TAddress;
  st: Pointer;
  Max: Longint;
 begin
  jt.link.GetAddress(Link, Address);

  jt.rests.Getst(restMaxRequestsBy, @Address, nil, st);

  if st = nil then
   MaxRequestedByCheck:=True
  else
   begin
    Counter:=0;

    ForEach(@Perform);

    Max:=Longint(jt.st.Get(st, 'Max'));

    MaxRequestedByCheck:=(Max = 0) or (Counter < Max);
   end;
 end;

function TForwardRequestQueue.MaxRequestedToCheck(const Link: Pointer): Boolean;
 var
  Counter: Longint;
 procedure Perform(const Item: PForwardRequestQueueItem); far;
  begin
   if Item^.RequestedFrom = Link then
    Inc(Counter);
  end;
 var
  Address: TAddress;
  st: Pointer;
  Max: Longint;
 begin
  jt.link.GetAddress(Link, Address);

  jt.rests.Getst(restMaxRequestsTo, @Address, nil, st);

  if st = nil then
   MaxRequestedToCheck:=True
  else
   begin
    Counter:=0;

    ForEach(@Perform);

    Max:=Longint(jt.st.Get(st, 'Max'));

    MaxRequestedToCheck:=(Max = 0) or (Counter < Max);
   end;
 end;

{ LoadQueue }

procedure LoadQueue;
 var
  Stream: PRamStream;
 begin
  Stream:=New(PRamStream, Init);

  if jt.rf.Get(fwdreqrf, ForwardRequestQueueID, Stream, StreamLinker) then
   begin
    Stream^.Seek(0);

    fwdreqQueue^.Load(Stream^);
   end;

  Dispose(Stream, Done);
 end;

{ StoreQueue }

procedure StoreQueue;
 var
  Stream: PRamStream;
 begin
  Stream:=New(PRamStream, Init);

  fwdreqQueue^.Store(Stream^);

  {$IFNDEF NOQUEUESTORE}
  jt.rf.Add(fwdreqrf, ForwardRequestQueueID, Stream, StreamLinker);
  {$ENDIF}

  Dispose(Stream, Done);
 end;

end.