unit LinkLink;

interface
uses
     Objects,

     PlugAPI,
     PlugSIDE,
     PlugUTIL,

     LinkAPI,

     jtMHLcmn,

     Wizard;

const
 Timeout = 100;

type
 TLinkerStatus = (lsOpening, lsReading, lsLinking, lsOrphanCheck, lsWriting, lsClosing);

 PLinkerMessages = ^TLinkerMessages;
 TLinkerMessages = object(TCollection)
  Disposer: TDisposeLoadLinkMessageBase;

  procedure FreeItem(Item: Pointer); virtual;
 end;

 PLinkerCache = ^TLinkerCache;
 TLinkerCache = object(TSortedCollection)
  Disposer: TDisposeLoadLinkMessageBase;

  function Compare(Key1, Key2: Pointer): xInteger; virtual;
  procedure FreeItem(Item: Pointer); virtual;
  function KeyOf(Item: Pointer): Pointer; virtual;
 end;

 PLinkerOrphaned = ^TLinkerOrphaned;
 TLinkerOrphaned = object(TLinkerMessages)
  procedure FreeItem(Item: Pointer); virtual;
 end;

 PLinker = ^TLinker;
 TLinker = object(TObject)
  constructor Init(const AArea: Pointer; const ACurrent, ATotal: Longint; const AInfo: Boolean);
  procedure Link;
  destructor Done; virtual;
 public
  function Status(const LinkerStatus: TLinkerStatus): Boolean;
 private
  Area, mb: Pointer;
  AreaName: String;
  Current, Total, No, TotalNo, Timer: Longint;
  PreviousLinkerStatus: TLinkerStatus;
  Info: Boolean;
 private
  LoadBlock: TLoadLinkMessageBaseBlock;
  StoreBlock: TStoreLinkMessageBaseBlock;
 private
  Root: PLinkerMessage;
  Messages: PLinkerMessages;
  Orphaned: PLinkerOrphaned;
  Cache: PLinkerCache;
 private
  procedure Dump(var Level: xInteger; const Message: PLinkerMessage; const DoNext: Boolean);
 end;

var
 Roller: TRoller;

const
 Debugger               : TDebugger = (Enabled: False; Prefix: 'Linker');
 DebuggerDump           : Boolean   = False;

function AddThunk(const Param: Pointer; const No, Global, MSGIDCRC, REPLYCRC: Longint;
  const Additional: Pointer): Boolean; export;
function StoreCallback(const Param: Pointer): Boolean; export;

implementation

{ AddThunk }

function AddThunk(const Param: Pointer; const No, Global, MSGIDCRC, REPLYCRC: Longint;
   const Additional: Pointer): Boolean;
 var
  Linker: PLinker absolute Param;
  Message: PLinkerMessage;
 begin
  New(Message);

  Message^.No:=No;
  Message^.Global:=Global;
  Message^.MSGIDCRC:=MSGIDCRC;
  Message^.REPLYCRC:=REPLYCRC;

  Message^.Owner:=nil;
  Message^.ChildFirst:=nil;
  Message^.ChildLast:=nil;
  Message^.ChildCount:=0;
  Message^.Next:=nil;

  Message^.Additional:=Additional;

  Linker^.Messages^.Insert(Message);

  Inc(Linker^.No);

  AddThunk:=not Linker^.Status(lsReading);
 end;

{ StoreCallback }

function StoreCallback(const Param: Pointer): Boolean;
 var
  Linker: PLinker absolute Param;
 begin
  Inc(Linker^.No);

  StoreCallback:=not Linker^.Status(lsWriting);
 end;

{ PreviousOf }

function PreviousOf(const Message: PLinkerMessage): PLinkerMessage;
 var
  M: PLinkerMessage;
 begin
  M:=Message^.Owner^.ChildFirst;

  while (M <> nil) and (M^.Next <> Message) do
   M:=M^.Next;

  PreviousOf:=M;
 end;

{ TLinkerMessages }

procedure TLinkerMessages.FreeItem(Item: Pointer);
 begin
  if @Disposer <> nil then
   Disposer(PLinkerMessage(Item)^.Additional);

  Dispose(PLinkerMessage(Item));
 end;

{ TLinkerCache }

function TLinkerCache.Compare(Key1, Key2: Pointer): xInteger;
 begin
  Compare:=CompareLongint(Longint(Key1), Longint(Key2));
{
  if Longint(Key1) < Longint(Key2) then Compare:=-1 else
  if Longint(Key1) > Longint(Key2) then Compare:=1 else
   Compare:=0;
}
 end;

procedure TLinkerCache.FreeItem(Item: Pointer);
 begin
 end;

function TLinkerCache.KeyOf(Item: Pointer): Pointer;
 begin
  KeyOf:=Pointer(PLinkerMessage(Item)^.MSGIDCRC);
 end;

{ TLinkerOrphaned }

procedure TLinkerOrphaned.FreeItem(Item: Pointer);
 begin
 end;

{ TLinker }

constructor TLinker.Init(const AArea: Pointer; const ACurrent, ATotal: Longint; const AInfo: Boolean);
 var
  rc: String;
  Need: Longint;
 begin
  inherited Init;

  Area:=AArea;

  GetPStringEx(jt.area.GetName(Area), AreaName);

  Current:=ACurrent;
  Total:=ATotal;

  Info:=AInfo;

  PreviousLinkerStatus:=lsWriting;

  Status(lsOpening);

  if not jt.area.Lock(Area, mb, rc) then
   begin
    jt.LogWrite('?', 'Unable to open ' + AreaName + ': ' + rc);

    Fail;
   end;

  if jt.mb.Count(mb) > MaxCollectionSize - 16 then
   begin
    jt.LogWrite('?', 'Unable to link more than ' + Long2Str(MaxCollectionSize - 16) + ' messages in ' + AreaName);

    jt.area.Unlock(Area, mb);

    Fail;
   end;

  Need:=jt.mb.Count(mb) * (SizeOf(TLinkerMessage) + 32) + 64 * 1024;

  if Need > MemAvail then
   begin
    jt.LogWrite('?', 'Too low memory for linking ' + AreaName + ' (need ' + Long2Str(Need) +
      ', but have only ' + Long2Str(MemAvail) + ')');

    jt.area.Unlock(Area, mb);

    Fail;
   end;

  if not jt.mb.Lock(mb) then
   begin
    jt.LogWrite('?', 'Unable to lock ' + AreaName + ' for purging');

    jt.area.Unlock(Area, mb);

    Fail;
   end;

  New(Root);

  FillChar(Root^, SizeOf(Root^), 0);

  Root^.MSGIDCRC:=-1;
  Root^.REPLYCRC:=-1;

  Root^.No:=-1;
  Root^.Global:=-1;

  Messages:=New(PLinkerMessages, Init(jt.mb.Count(mb), $100));

  Orphaned:=New(PLinkerOrphaned, Init(jt.mb.Count(mb), $100));

  Cache:=New(PLinkerCache, Init(jt.mb.Count(mb), $100));

  Cache^.Duplicates:=True;
 end;

procedure TLinker.Link;
 procedure Add(const Message: PLinkerMessage; Target: PLinkerMessage; const AddToCache: Boolean);
  var
   Index: xInteger;
  begin
   Cache^.Search(Pointer(Message^.MSGIDCRC), Index);

   if Target <> Root then
    if LoadBlock.Flags and llmbbfMsgLike <> 0 then
     begin
      while Target^.ChildFirst <> nil do
       Target:=Target^.ChildFirst;

      if Debugger.Enabled then
       Debugger.Log(AreaName + ': msglike for ' + Long2Str(Message^.No) + ', target=' + HexPtr(Target));
     end
    else
     if (LoadBlock.Flags and llmbbfBranchLimit <> 0) and (Target^.ChildCount >= LoadBlock.BranchLimit) then
      begin
       while Target^.ChildCount >= LoadBlock.BranchLimit do
        Target:=Target^.ChildLast;

       if Debugger.Enabled then
        Debugger.Log(AreaName + ': branchlimit for ' + Long2Str(Message^.No) + ', target=' + HexPtr(Target));
      end;

   if (Target = Root) and (Message^.MSGIDCRC <> -1) then
    Orphaned^.Insert(Message);

   Message^.Owner:=Target;

   if Target^.ChildFirst = nil then
    begin
     Target^.ChildFirst:=Message;
     Target^.ChildLast:=Message;

     Inc(Target^.ChildCount);
    end
   else
    begin
     Target^.ChildLast^.Next:=Message;
     Target^.ChildLast:=Message;

     Inc(Target^.ChildCount);
    end;

   if AddToCache then
    Cache^.AtInsert(Index, Message);

   if Debugger.Enabled then
    Debugger.Log(AreaName + ': ins ' + Long2str(Message^.No) + ' to ' + Long2Str(Target^.No));
  end;
 function PerformLink(const Message: PLinkerMessage): Boolean; far;
  var
   Index: xInteger;
  begin
   Inc(No);

   if (Message^.MSGIDCRC = -1) or (Message^.REPLYCRC = -1) or not Cache^.Search(Pointer(Message^.REPLYCRC), Index) then
    Add(Message, Root, True)
   else
    Add(Message, Cache^.At(Index), True);

   PerformLink:=Status(lsLinking);
  end;
 var
  Message, Target, Previous: PLinkerMessage;
  Index: xInteger;
  K, ReallyOrphaned, GlobalTimer, ChainsCount, OrphanedCount: Longint;
 begin
  GlobalTimer:=Clock;

  LoadBlock.Param:=@Self;
  LoadBlock.Add:=AddThunk;
  LoadBlock.Area:=Area;

  TotalNo:=jt.mb.Count(mb) * 3;
  No:=0;

  jt.mb.LoadLinkInfo(mb, LoadBlock);

  Messages^.Disposer:=LoadBlock.Dispose;
  Cache^.Disposer:=LoadBlock.Dispose;

  if not LoadBlock.Supported then
   begin
    jt.LogWrite('?', AreaName + ': linking is not supported');

    Exit;
   end;

  if not LoadBlock.Did then
   Exit;

  if Messages^.FirstThat(@PerformLink) = nil then
   begin
    Status(lsOrphanCheck);

    Timer:=Clock;

    ReallyOrphaned:=0;

    while ReallyOrphaned <> Orphaned^.Count do
     begin
      ReallyOrphaned:=0;
      K:=0;

      if Clock - Timer > 500 then
       begin
        Debugger.Log(AreaName + ': orphan-check timeout exceeded');

        Break;
       end;

      while K < Orphaned^.Count do
       begin
        Message:=Orphaned^.At(K);

        if (Message^.REPLYCRC <> -1) and Cache^.Search(Pointer(Message^.REPLYCRC), Index) then
         begin
          if Message^.Owner = nil then
           jt.LogWrite('~', AreaName + ': failed while orphan-check, I''ll crash here, my brain is dead');

          Target:=Cache^.At(Index);

          if Message <> Target then
           begin
            if Message = Message^.Owner^.ChildFirst then
             Message^.Owner^.ChildFirst:=Message^.Next;

            Previous:=PreviousOf(Message);

            if Debugger.Enabled then
             if Previous = nil then
              Debugger.Log(AreaName + ': ' + Long2Str(Message^.No) + ', prev=nil')
             else
              Debugger.Log(AreaName + ': ' + Long2Str(Message^.No) +
                                      ', prev=' + HexPtr(Previous) + ', prev-next=' +
                                                  HexPtr(Previous^.Next) + ', this=' +
                                                  HexPtr(Message));

            if Message = Message^.Owner^.ChildLast then
             Message^.Owner^.ChildLast:=Previous;

            if Previous <> nil then
             Previous^.Next:=Message^.Next;

            Message^.Next:=nil;

            Dec(Message^.Owner^.ChildCount);

            Add(Message, Target, False);

            if Debugger.Enabled then
             Debugger.Log(AreaName + ': ' + Long2Str(Message^.No) + ' must moved to ' + Long2Str(Target^.No));

            Orphaned^.AtDelete(K);
           end
          else
           begin
            Inc(K);

            Inc(ReallyOrphaned);
           end;
         end
        else
         begin
          Inc(K);

          Inc(ReallyOrphaned);
         end;
       end;
     end;

    StoreBlock.Count:=Messages^.Count;
    StoreBlock.Items:=Messages^.Items;
    StoreBlock.Root:=Root;
    StoreBlock.Callback:=StoreCallback;
    StoreBlock.Param:=@Self;

    jt.mb.StoreLinkInfo(mb, StoreBlock);

    if Info then
     begin
      GlobalTimer:=Clock - GlobalTimer;

      ChainsCount:=0;
      OrphanedCount:=0;

      Message:=Root^.ChildFirst;

      while Message <> nil do
       begin
        if Message^.ChildFirst <> nil then
         Inc(ChainsCount)
        else
         Inc(OrphanedCount);

        Message:=Message^.Next;
       end;

      if (GlobalTimer > 0) and (ChainsCount > 0) then
       jt.LogWrite('L', AreaName + ': ' +
                        Long2Str(ChainsCount) + ' chain' + _s(ChainsCount) + ', ' +
                        Long2Str(Messages^.Count) + ' msg' + _s(Messages^.Count) + ', ' +
                        Long2Str(Messages^.Count - OrphanedCount) + ' in chains, ' +
                        Real2Str(Messages^.Count / (GlobalTimer / 1000), 1, 0) + ' msgs/s');
     end;

    if DebuggerDump then
     begin
      jt.LogWrite('~', 'Here follows dump of ' + AreaName + ' linked chains:');

      Index:=0;

      Dump(Index, Root, True);
     end;
   end;
 end;

destructor TLinker.Done;
 begin
  Status(lsClosing);

  Cache^.DeleteAll;

  Dispose(Cache, Done);

  Dispose(Orphaned, Done);

  Dispose(Messages, Done);

  Dispose(Root);

  jt.mb.Unlock(mb);

  jt.area.Unlock(Area, mb);

  inherited Done;
 end;

{ public methods }

function TLinker.Status(const LinkerStatus: TLinkerStatus): Boolean;
 var
  Percent: String;
 begin
  if (LinkerStatus <> PreviousLinkerStatus) or (Clock - Timer > Timeout) then
   begin
    Status:=jt.Cancelled;

    Percent:=jt.Percent(Current, Total);

    case LinkerStatus of
     lsOpening: jt.SetTitle(AreaName + ' ' + Percent + ' (opening)');
     lsReading: jt.SetTitle(AreaName + ' ' + Percent + ' (reading)');
     lsLinking: jt.SetTitle(AreaName + ' ' + Percent + ' (linking)');
     lsOrphanCheck: jt.SetTitle(AreaName + ' ' + Percent + ' (orphan-check)');
     lsWriting: jt.SetTitle(AreaName + ' ' + Percent + ' (writing)');
     lsClosing: jt.SetTitle(AreaName + ' ' + Percent + ' (closing)');
    end;

    case LinkerStatus of
     lsOpening, lsClosing: jt.SetStatus(AreaName + ' ' + Percent);
     lsReading: jt.SetStatus(AreaName + ' ' + Percent + ' ' + Roller.GetRoller + ' reading ' + jt.Percent(No, TotalNo));
     lsLinking: jt.SetStatus(AreaName + ' ' + Percent + ' ' + Roller.GetRoller + ' linking ' + jt.Percent(No, TotalNo));
     lsOrphanCheck: jt.SetStatus(AreaName + ' ' + Percent + ' ' + Roller.GetRoller + ' orphan-check ' +
       jt.Percent(No, TotalNo));
     lsWriting: jt.SetStatus(AreaName + ' ' + Percent + ' ' + Roller.GetRoller + ' writing ' + jt.Percent(No, TotalNo));
    end;

    PreviousLinkerStatus:=LinkerStatus;
    Timer:=Clock;
   end
  else
   Status:=False;
 end;

{ private methods }

procedure Dump_(const Level: xInteger; const Message: PLinkerMessage);
 begin
  jt.LogWrite('~', PadCh(Pad('', Level * 2) + ' ' +
                         Long2Str(Message^.Global) + ' ', '.', 30) +
                         ' ' + HexL(Message^.MSGIDCRC) +
                         ' ' + HexL(Message^.REPLYCRC) +
                         ' (' + Long2Str(Message^.ChildCount) + ')');
 end;

procedure TLinker.Dump(var Level: xInteger; const Message: PLinkerMessage; const DoNext: Boolean);
 var
  M: PLinkerMessage;
 begin
  Dump_(Level, Message);

  if Message^.ChildFirst <> nil then
   begin
    Inc(Level);

    Dump(Level, Message^.ChildFirst, True);

    Dec(Level);
   end;

  if DoNext then
   begin
    M:=Message^.Next;

    while M <> nil do
     begin
      Dump(Level, M, False);

      M:=M^.Next;
     end;
   end;
 end;

end.