library Rescan;

uses
     PlugAPI,
     PlugSIDE,
     PlugUTIL,

     Objects,

     TossAPI,
     TossCmn,

     afxAPI,
     afxRight,

     jtMHLcmn,

     Wizard;

{
  originally (c) by Alexander Bo, 2:5020/2388@fidonet, 2001.
  portions (c) by sergey korowkin, 2:6033/27@fidonet, 2001.

  Areafix command syntax:

  %RESCAN areas switches

  areas:

    area name (RU.FASTUUE) or areas wildcard (RU.FAST*)

  switches:

    days:number
      send messages arrived at <number> of lately days

    maxmsgs:number
      limit total messages per area

    date:date
      send messages written at <date>

    period:date-date
      send messages written at period of <date> till <date>

  <date> syntax: dd/mm/yyyy

  examples:

  %RESCAN RU.FASTUUE days:7
  %RESCAN RU.FASTUUE date:07/01/1983
  %RESCAN RU.FASTUUE period:01/01/2001-31/12-2001

  please use this source as proposal for writing Areafix command handlers
}

type
 TCriterionResult = (crApplicable, crNotApplicable, crStop);

 PCriterion = ^TCriterion;
 TCriterion = object(TObject)
  function Applicable(const msg: Pointer; msgDate: Longint): TCriterionResult; virtual;
  procedure Did(const msg: Pointer; msgDate: Longint); virtual;
 end;

 PDaysCriterion = ^TDaysCriterion;
 TDaysCriterion = object(TCriterion)
  Today, Days: Longint;

  constructor Init(const ADays: Longint);
  function Applicable(const msg: Pointer; msgDate: Longint): TCriterionResult; virtual;
 end;

 PMessagesCriterion = ^TMessagesCriterion;
 TMessagesCriterion = object(TCriterion)
  Now, Maximum: Longint;

  constructor Init(const AMaximum: Longint);
  function Applicable(const msg: Pointer; msgDate: Longint): TCriterionResult; virtual;
  procedure Did(const msg: Pointer; msgDate: Longint); virtual;
 end;

 PDateCriterion = ^TDateCriterion;
 TDateCriterion = object(TCriterion)
  Date: Longint;

  constructor Init(const ADate: Longint);
  function Applicable(const msg: Pointer; msgDate: Longint): TCriterionResult; virtual;
 end;

 PPeriodCriterion = ^TPeriodCriterion;
 TPeriodCriterion = object(TCriterion)
  StartPoint, EndPoint: Longint;

  constructor Init(const AStartPoint, AEndPoint: Longint);
  function Applicable(const msg: Pointer; msgDate: Longint): TCriterionResult; virtual;
 end;

 PCriterions = ^TCriterions;
 TCriterions = object(TCollection)
  function Applicable(const msg: Pointer; msgDate: Longint): TCriterionResult; virtual;
  procedure Did(const msg: Pointer; msgDate: Longint); virtual;
 end;

const
 RescannerPlugin                : Pointer       = nil;
 RescannerAddRescanHeader       : Boolean       = True;
 RescannerNoRescan              : Pointer       = nil;
 RescannerMaxMessages           : Longint       = 1024;

{ TCriterion }

function TCriterion.Applicable(const msg: Pointer; msgDate: Longint): TCriterionResult;
 begin
  Abstract;
 end;

procedure TCriterion.Did(const msg: Pointer; msgDate: Longint);
 begin
 end;

{ TDaysCriterion }

constructor TDaysCriterion.Init(const ADays: Longint);
 var
  DT: TDateTime;
 begin
  inherited Init;

  Days:=ADays;

  GetCurrentDateTime(DT);

  Today:=GregorianToJulian(DT);
 end;

function TDaysCriterion.Applicable(const msg: Pointer; msgDate: Longint): TCriterionResult;
 begin
  if (Today - msgDate <= Days) and (Today - msgDate >= 0) then
   Applicable:=crApplicable
  else
   Applicable:=crNotApplicable;
 end;

{ TMessagesCriterion }

constructor TMessagesCriterion.Init(const AMaximum: Longint);
 begin
  inherited Init;

  Maximum:=AMaximum;
  Now:=0;
 end;

function TMessagesCriterion.Applicable(const msg: Pointer; msgDate: Longint): TCriterionResult;
 begin
  if Now < Maximum then
   Applicable:=crApplicable
  else
   Applicable:=crStop;
 end;

procedure TMessagesCriterion.Did(const msg: Pointer; msgDate: Longint);
 begin
  Inc(Now);
 end;

{ TDateCriterion }

constructor TDateCriterion.Init(const ADate: Longint);
 begin
  inherited Init;

  Date:=ADate;
 end;

function TDateCriterion.Applicable(const msg: Pointer; msgDate: Longint): TCriterionResult;
 var
  DT: TDateTime;
 begin
  jt.msg.GetWritten(msg, DT);

  if GregorianToJulian(DT) = Date then
   Applicable:=crApplicable
  else
   Applicable:=crNotApplicable;
 end;

{ TPeriodCriterion }

constructor TPeriodCriterion.Init(const AStartPoint, AEndPoint: Longint);
 begin
  inherited Init;

  StartPoint:=AStartPoint;
  EndPoint:=AEndPoint;
 end;

function TPeriodCriterion.Applicable(const msg: Pointer; msgDate: Longint): TCriterionResult;
 var
  DT: TDateTime;
 begin
  jt.msg.GetWritten(msg, DT);

  msgDate:=GregorianToJulian(DT);

  if (msgDate >= StartPoint) and (msgDate <= EndPoint) then
   Applicable:=crApplicable
  else
   Applicable:=crNotApplicable;
 end;

{ TCriterions }

function TCriterions.Applicable(const msg: Pointer; msgDate: Longint): TCriterionResult;
 function Perform(const Criterion: PCriterion): Boolean; far;
  var
   CriterionResult: TCriterionResult;
  begin
   CriterionResult:=Criterion^.Applicable(msg, msgDate);

   if CriterionResult in [crStop, crNotApplicable] then
    begin
     Perform:=True;

     Applicable:=CriterionResult;
    end
   else
    Perform:=False;
  end;
 begin
  if FirstThat(@Perform) = nil then
   Applicable:=crApplicable;
 end;

procedure TCriterions.Did(const msg: Pointer; msgDate: Longint);
 procedure Perform(const Criterion: PCriterion); far;
  begin
   Criterion^.Did(msg, msgDate);
  end;
 begin
  ForEach(@Perform);
 end;

{ RescanHandler }

procedure RescanHandler(const Command: String; Value: String); export;
 var
  Reply: Pointer;
 procedure Error(const Message: String);
  begin
   jt.msg.Add(Reply, Trim('%RESCAN ' + Value) + ': ' + Message);

   jt.LogWrite('?', Message);
  end;
 procedure InternalError(const Message: String);
  begin
   jt.LogWrite('?', Message);
  end;
 function ParseDate(const Switch: String; var Date: Longint; var rc: String): Boolean;
  var
   DT: TDateTime;
  begin
   ParseDate:=False;

   if not Str2Word(ExtractWord(1, Switch, ['/']), DT.Day) then rc:='failed to parse day of the date' else
   if not Str2Word(ExtractWord(2, Switch, ['/']), DT.Month) then rc:='failed to parse month of the date' else
   if not Str2Word(ExtractWord(3, Switch, ['/']), DT.Year) then rc:='failed to parse year of the date' else
    begin
     Date:=GregorianToJulian(DT);

     if not IsValidDateTime(DT) then
      rc:='invalid date ' + lz(DT.Day) + '/' +
                            lz(DT.Month) + '/' +
                            Long2Str(DT.Year)
     else
      ParseDate:=True;
    end;
  end;
 var
  AreasRescanned, MessagesSent, ThisSent, TotalMessages: Longint;
  I, K, Divider, Number, Number2, Timer: Longint;
  AreaName, AreaMask, Switch, rc: String;
  Area, LinkArea, mb, msg: Pointer;
  LinkAddress: TAddress;
  Criterions: PCriterions;
  Created: Boolean;
  Block: TAreafixGetRequestInfoBlock;
  List: PNumbers;
  DT: TDateTime;
 function RescanMessage(const No: Longint): Boolean; far;
  var
   msgst: Pointer;
   FromAddress, ToAddress: TAddress;
   Password: String;
   TossMessageToLinkDistinguishHostOMLBlock: TTossMessageToLinkDistinguishHostOMLBlock;
  begin
   RescanMessage:=False;

   Inc(I);

   if not jt.mb.Load(mb, No, msg, rc, mfNone) then
    Exit;

   jt.msg.SetArea(msg, AreaName);

   msgst:=jt.msg.st(msg);

   jt.link.GetAka(Block.Link, FromAddress);
   jt.st.Add(msgst, 'Packet.From', jt.NewAddress(FromAddress), jt.svc.stAddress);

   jt.link.GetAddress(Block.Link, ToAddress);
   jt.st.Add(msgst, 'Packet.To', jt.NewAddress(ToAddress), jt.svc.stAddress);

   jt.link.GetOutPassword(Block.Link, Password);
   jt.st.Add(msgst, 'Packet.Password', jt.NewStr(Password), jt.svc.stString);

   if RescannerAddRescanHeader then
    jt.msg.AddAfterKludges(msg, '*** rescanned at ' + AddressToStrEx(FromAddress) +
                                                      jt.JokeSwitch(' for ', ' specially for ') +
                                                      AddressToStrEx(ToAddress) +
                                                      jt.JokeSwitch('', ' ;-))'));

   TossMessageToLinkDistinguishHostOMLBlock.Link:=Block.Link;
   TossMessageToLinkDistinguishHostOMLBlock.msg:=msg;
   TossMessageToLinkDistinguishHostOMLBlock.Host:=@DefaultEchomailSessionHost;
   TossMessageToLinkDistinguishHostOMLBlock.oml:=jt.link.oml(Block.Link);

   if Clock - Timer > 100 then
    begin
     jt.SetStatus('%RESCAN ' + AreaName + ' ' + jt.Percent(K, jt.link.area.Count(Block.Link)) +
                  ' sending to ' + AddressToStrEx(ToAddress) + ' ' +
                  jt.Percent(I, List^.Count));

     Timer:=Clock;

     if jt.Cancelled then
      RescanMessage:=True;
    end;

   with TossMessageToLinkDistinguishHostOMLBlock do
    RecycleLink(oml, Block.Link, msg, Host^, FromAddress, ToAddress, Password);

   Inc(ThisSent);

   jt.msg.Destroy(msg);
  end;
 begin
  Reply:=afxCreateReply('Rescan.' + GenerateMSGID + '.' + Value, Created);

  if Created then
   begin
    jt.msg.SetSubject(Reply, Trim('%RESCAN ' + Value) + ' results');

    afxFixEmptyReply(Reply);
   end;

  ReplaceChar(Value, #9, #32);

  AreaMask:=ExtractWord(1, Value, [' ']);

  StUpcaseEx(AreaMask);

  Criterions:=New(PCriterions, Init($10, $10));

  for K:=2 to WordCount(Value, [' ']) do
   begin
    Switch:=ExtractWord(K, Value, [' ']);

    StUpcaseEx(Switch);

    if (Switch <> '') and (Switch[1] in ['/', '-']) then
     Delete(Switch, 1, 1);

    if Switch = '' then
     Continue;

    repeat
     if Copy(Switch, 1, 5) = 'DAYS:' then
      begin
       Delete(Switch, 1, 5);

       if not Str2Longint(Switch, Number) then Error('unable to parse DAYS value "' + Switch + '"') else
       if Number < 0 then Error('invalid DAYS value "' + Switch + '"') else
        begin
         Criterions^.Insert(New(PDaysCriterion, Init(Number)));

         Continue;
        end;
      end else
     if Copy(Switch, 1, 8) = 'MAXMSGS:' then
      begin
       Delete(Switch, 1, 8);

       if not Str2Longint(Switch, Number) then Error('unable to parse MAXMSGS value "' + Switch + '"') else
       if Number < 0 then Error('invalid MAXMSGS value "' + Switch + '"') else
        begin
         Criterions^.Insert(New(PMessagesCriterion, Init(Number)));

         Continue;
        end;
      end else
     if Copy(Switch, 1, 5) = 'DATE:' then
      begin
       Delete(Switch, 1, 5);

       if not ParseDate(Switch, Number, rc) then Error('unable to parse DATE value "' + Switch + '": ' + rc) else
        begin
         Criterions^.Insert(New(PDateCriterion, Init(Number)));

         Continue;
        end;
      end else
     if Copy(Switch, 1, 7) = 'PERIOD:' then
      begin
       Delete(Switch, 1, 7);

       Divider:=Pos('-', Switch);

       if Divider = 0 then Error('unable to parse PERIOD:' + Switch + ': "-" must be specified to separate dates') else
       if not ParseDate(Copy(Switch, 1, Divider - 1), Number, rc) then Error('unable to parse PERIOD:' + Switch +
         ' start-date: ' + rc) else
       if not ParseDate(Copy(Switch, Divider + 1, 255), Number2, rc) then Error('unable to parse PERIOD:' + Switch +
         ' end-date: ' + rc) else
       if Number > Number2 then Error('unable to parse PERIOD:' + Switch + ' value: start-date larger than end-date') else
        begin
         Criterions^.Insert(New(PPeriodCriterion, Init(Number, Number2)));

         Continue;
        end;
      end
     else
      Error('unknown switch "' + Switch + '"');

     Dispose(Criterions, Done);

     Exit;
    until True;
   end;

  if Criterions^.Count = 0 then
   begin
    Error('no criterions specified, use %HELP to learn more about %RESCAN syntax');

    Dispose(Criterions, Done);

    Exit;
   end;

  Criterions^.Insert(New(PMessagesCriterion, Init(RescannerMaxMessages)));

  afxGetRequestInfo(Block);

  jt.link.GetAddress(Block.Link, LinkAddress);

  AreasRescanned:=0;
  MessagesSent:=0;

  for K:=0 to jt.link.area.Count(Block.Link) - 1 do
   begin
    LinkArea:=jt.link.area.At(Block.Link, K);

    Area:=jt.link.area.GetArea(LinkArea);

    AreaName:=jt.area.GetName(Area)^;

    if CheckWildcard(AreaName, AreaMask) and
       (not jt.link.area.IsPassive(LinkArea)) and
       jt.link.area.IsSubscribed(LinkArea) and
       (not jt.area.Passthrough(Area)) and
       (not jt.rests.Check(RescannerNoRescan, @LinkAddress, @AreaName)) and
       afxIsOperable(Area) then
     begin
      jt.SetStatus('%RESCAN ' + AreaName + ' opening');

      if jt.area.Lock(Area, mb, rc) then
       begin
        if jt.mb.Lock(mb) then
         begin
          List:=New(PNumbers, Init(jt.mb.Count(mb), $100));

          Timer:=0;

          TotalMessages:=jt.mb.Count(mb);

          for I:=TotalMessages - 1 downto 0 do
           begin
            if Clock - Timer > 100 then
             begin
              jt.SetStatus('%RESCAN ' + AreaName + ' ' + jt.Percent(K, jt.link.area.Count(Block.Link)) +
                             ' scanning ' + jt.Percent(TotalMessages - I, TotalMessages));

              Timer:=Clock;

              if jt.Cancelled then
               Break;
             end;

            if jt.mb.Load(mb, I, msg, rc, mfNone) then
             begin
              jt.msg.GetArrived(msg, DT);

              if not IsValidDateTime(DT) then
               jt.msg.GetWritten(msg, DT);

              Number:=GregorianToJulian(DT);

              case Criterions^.Applicable(msg, Number) of
               crApplicable: List^.Add(I);
               crStop: Break;
              end;

              jt.msg.Destroy(msg);
             end
            else
             jt.LogWrite('?', 'Failed to open msg ' + Long2Str(I) + ' in ' + AreaName + ': ' + rc);
           end;

          if (List^.Count <> 0) and not jt.Cancelled then
           begin
            ThisSent:=0;

            Timer:=0;

            I:=0;

            List^.FirstThat(@RescanMessage);

            if ThisSent <> 0 then
             begin
              Inc(AreasRescanned);
              Inc(MessagesSent, ThisSent);

              jt.LogWrite('A', '%RESCAN ' + AreaName + ': ' + Long2StrFmt(ThisSent) + ' message' + _s(ThisSent) + ' sent');

              jt.msg.Add(Reply, AreaName + ': ' + Long2StrFmt(ThisSent) + ' message' + _s(ThisSent) + ' sent');
             end;
           end;

          Dispose(List, Done);
         end
        else
         InternalError('failed to lock ' + AreaName);

        jt.area.Unlock(Area, mb);
       end
      else
       InternalError('failed to open ' + AreaName + ': ' + rc);

      jt.ClearStatus;
     end;

    if jt.Cancelled then
     Break;
   end;

  Dispose(Criterions, Done);

  jt.ClearStatus;

  if AreasRescanned <> 0 then
   begin
    jt.LogWrite('A', 'Total ' + Long2StrFmt(AreasRescanned) +
                      ' area' + _s(AreasRescanned) +
               ' rescanned, ' + Long2StrFmt(MessagesSent) +
                   ' message' + _s(MessagesSent) + ' sent');

    jt.msg.Add(Reply, 'Total ' + Long2StrFmt(AreasRescanned) +
                       ' area' + _s(AreasRescanned) +
                ' rescanned, ' + Long2StrFmt(MessagesSent) +
                    ' message' + _s(MessagesSent) + ' sent');
   end
  else
   begin
    jt.LogWrite('A', 'Nothing to rescan');

    jt.msg.Add(Reply, 'Nothing to rescan');
   end;
 end;

{ PluginSetup }

function PluginSetup(const JTAPI: Pointer; const PlugInfo: PPluginInformation; const Plugin: Pointer): Boolean; export;
 begin
  SetupAPI(JTAPI);

  SetupPlugInfo(PlugInfo, 'Rescanner', 1);

  RescannerPlugin:=Plugin;

  PluginSetup:=True;

  RescannerNoRescan:=jt.rests.Add('afxNoRESCAN', rsMinus);
 end;

{ PluginStartup }

procedure PluginStartup; export;
 begin
  afxRegisterCommandHandler('RESCAN', RescanHandler);
 end;

{ PluginHandleConfigSwitch }

procedure PluginHandleConfigSwitch(const Keyword: String; Value: String; const Linker: Pointer); export;
 begin
  if Keyword = 'ADDRESCANHEADER' then RescannerAddRescanHeader:=jt.linker.ParseBoolean(Linker, Value) else
  if Keyword = 'MAXMESSAGES' then
   begin
    RescannerMaxMessages:=jt.linker.ParseNumber(Linker, Value);

    if RescannerMaxMessages < 16 then
     jt.linker.Error(Linker, 'MaxMessages must be greater than 15');
   end else
   jt.linker.Error(Linker, 'syntax error');
 end;

exports
 PluginSetup,
 PluginStartup,
 PluginHandleConfigSwitch;

begin
 Assign(Output, '');
 Rewrite(Output);
end.