unit ResFile;

interface
uses
     PlugSIDE,

{$IFNDEF SINGLE}
     HostAPI,
{$ELSE}
     Drivers,
{$ENDIF}

     Thunks,

     jtMHLstm,

     Objects,

     RStream,

     Wizard;

type
 PResErrors = ^TResErrors;
 TResErrors = object(TCollection)
  procedure FreeItem(Item: Pointer); virtual;
 end;

 PResItem = ^TResItem;
 TResItem = object(TObject)
  HeaderOffset: Longint;
  Name: PString;
  DataOffset: Longint;
  DataSize: Longint;
  Size: Longint;
  Junk: Boolean;

  constructor Init(const AHeaderOffset: Longint; const AName: PString);
  destructor Done; virtual;
 public
  procedure UpdateHeader(const Stream: PStream);
  procedure UpdateHeaderIndirect(const Stream: PStream);
  procedure CalculateDataOffset;
 end;

 PResIndex = ^TResIndex;
 TResIndex = object(TCollection)
  function Seek(const Name: String): PResItem;
  function SeekJunk(const Size: Longint): PResItem;
 end;

 PResFile = ^TResFile;
 TResFile = object(TObject)
  Index: PResIndex;
  Errors: PResErrors;
  Stream: PStream;
  FName: PString;
  Flags, UsedCount: Longint;

  constructor Init(const AFName: String; const AFlags: Longint);
  destructor Done; virtual;
 public
  procedure AddError(const S: String);
  function Error: Boolean;
  procedure Flush;
  procedure GetSizes(var JunkSize, DataSize, TotalSize: Longint);
  function CalculateItemSize(const Name: String; const DataSize: Longint): Longint;
  function Add(const Name: String; const S: PStream): Boolean;
  function Remove(const Name: String): Boolean;
  function Get(const Name: String; const S: PStream): Boolean;
  function Pack(const TempStream: PStream): Boolean;
  function AddExternal(const Name: String; const S: Pointer; const Linker: PStreamLinker): Boolean;
  function GetExternal(const Name: String; const S: Pointer; const Linker: PStreamLinker): Boolean;
  function PackExternal(const S: Pointer; const Linker: PStreamLinker): Boolean;
  function Count: Longint;
  function At(const I: Longint; var Junk: Boolean; var Size: Longint): PString;
 end;

 PResFiles = ^TResFiles;
 TResFiles = object(TSortedCollection)
  function Compare(Key1, Key2: Pointer): xInteger; virtual;
  function KeyOf(Item: Pointer): Pointer; virtual;
 end;

const
 ResFileItemID        : Longint         = $DABAC00A;

 ResFiles             : PResFiles       = nil;

 ResFilesForcePack    : Boolean         = False;

procedure rfInit;
procedure rfDone;

implementation
uses
     PlugAPI;

{$IFDEF SINGLE}
procedure LogWrite(const Ch: Char; const S: String);
 begin
  PrintStr(S + #13#10);
 end;
{$ENDIF}

{ TResErrors }

procedure TResErrors.FreeItem(Item: Pointer);
 begin
  DisposeStr(Item);
 end;

{ TResItem }

constructor TResItem.Init(const AHeaderOffset: Longint; const AName: PString);
 begin
  inherited Init;

  HeaderOffset:=AHeaderOffset;

  Name:=AName;
 end;

destructor TResItem.Done;
 begin
  DisposeStr(Name);

  inherited Done;
 end;

{ public methods }

procedure TResItem.UpdateHeader(const Stream: PStream);
 begin
  Stream^.Seek(HeaderOffset);

  UpdateHeaderIndirect(Stream);
 end;

procedure TResItem.UpdateHeaderIndirect(const Stream: PStream);
 begin
  Stream^.Write(ResFileItemID, SizeOf(ResFileItemID));

  Stream^.WriteStr(Name);

  Stream^.Write(DataSize, SizeOf(DataSize));
  Stream^.Write(Size, SizeOf(Size));

  Stream^.Write(Junk, SizeOf(Junk));
 end;

procedure TResItem.CalculateDataOffset;
 begin
  DataOffset:=HeaderOffset +
              SizeOf(ResFileItemID) +
              Length(Name^) + 1 +
              SizeOf(DataSize) +
              SizeOf(Size) +
              SizeOf(Junk);
 end;

{ TResIndex }

function TResIndex.Seek(const Name: String): PResItem;
 function Lookup(const Item: PResItem): Boolean; far;
  begin
   Lookup:=(Item^.Name^ = Name) and not Item^.Junk;
  end;
 begin
  Seek:=FirstThat(@Lookup);
 end;

function TResIndex.SeekJunk(const Size: Longint): PResItem;
 function Lookup(const Item: PResItem): Boolean; far;
  begin
   Lookup:=(Item^.Size >= Size) and Item^.Junk;
  end;
 begin
  SeekJunk:=FirstThat(@Lookup);
 end;

{ TResFile }

constructor TResFile.Init(const AFName: String; const AFlags: Longint);
 var
  Name: PString;
  Item: PResItem;
  Pos, Size, id: Longint;
 begin
  inherited Init;

  Flags:=AFlags;

  Index:=New(PResIndex, Init($10, $10));
  Errors:=New(PResErrors, Init($0, $4));
  Stream:=New(PDosStreamEx, Init(AFName, stOpen));

  if Stream^.Status <> stOk then
   begin
    Dispose(Stream, Done);

    Stream:=New(PDosStream, Init(AFName, stCreate));

    if Stream^.Status <> stOk then
     begin
      Dispose(Stream, Done);

      Fail;
     end;
   end
  else
   begin
    Pos:=Stream^.GetPos;
    Size:=Stream^.GetSize;

    while Pos < Size do
     begin
      Stream^.Read(id, SizeOf(id));

      if id <> ResFileItemID then
       begin
        AddError('unknown id at ' + HexL(Pos) + ': got ' + HexL(id) + ', expected ' + HexL(ResFileItemID));

        Break;
       end;

      Item:=New(PResItem, Init(Pos, Stream^.ReadStr));

      Stream^.Read(Item^.DataSize, SizeOf(Item^.DataSize));
      Stream^.Read(Item^.Size, SizeOf(Item^.Size));
      Stream^.Read(Item^.Junk, SizeOf(Item^.Junk));

      Item^.DataOffset:=Stream^.GetPos;

      Stream^.Seek(Item^.HeaderOffset + Item^.Size);

      if Stream^.Status <> stOk then
       begin
        id:=Stream^.Status;

        Stream^.Reset;

        AddError('stream error ' + HexL(id) + ' at frame ' + HexL(Pos) + ' near offset ' + HexL(Stream^.GetPos));

        Dispose(Item, Done);

        Break;
       end;

      Index^.Insert(Item);

      Pos:=Stream^.GetPos;
     end;
   end;

  FName:=NewStr(AFName);
 end;

destructor TResFile.Done;
 var
  JunkSize, DataSize, TotalSize: Longint;
  TempStream: PStream;
 procedure DumpError(const S: PString); far;
  begin
   LogWrite('!', S^);
  end;
 begin
  if (Flags and rffManualPack = 0) or (Flags and rffForcedPack <> 0) or ResFilesForcePack then
   begin
    GetSizes(JunkSize, DataSize, TotalSize);

    if (Flags and rffForcedPack <> 0) or ResFilesForcePack or
       ((TotalSize <> 0) and (JunkSize > TotalSize div 3) and (MemAvail > DataSize * 2)) then
     begin
      TempStream:=New(PRamStream, Init);

      Pack(TempStream);

      Dispose(TempStream, Done);
     end;
   end;

  if Errors^.Count <> 0 then
   begin
    LogWrite('!', 'Resource file ' + FName^ + ' has errors:');

    Errors^.ForEach(@DumpError);
   end;

  Dispose(Stream, Done);
  Dispose(Errors, Done);
  Dispose(Index, Done);

  DisposeStr(FName);

  inherited Done;
 end;

{ public methods }

procedure TResFile.AddError(const S: String);
 begin
  if S <> '' then
   Errors^.Insert(NewStr(S));
 end;

function TResFile.Error: Boolean;
 begin
  Error:=Errors^.Count <> 0;
 end;

procedure TResFile.Flush;
 begin
  if Stream^.Status = stOk then
   Stream^.Flush;
 end;

procedure TResFile.GetSizes(var JunkSize, DataSize, TotalSize: Longint);
 procedure CountSizes(const Item: PResItem); far;
  begin
   Inc(JunkSize, Item^.Size - Item^.DataSize - CalculateItemSize(Item^.Name^, 0));
   Inc(DataSize, Item^.DataSize);
   Inc(TotalSize, Item^.Size);
  end;
 begin
  JunkSize:=0;
  DataSize:=0;
  TotalSize:=0;

  Index^.ForEach(@CountSizes);
 end;

function TResFile.CalculateItemSize(const Name: String; const DataSize: Longint): Longint;
 begin
  CalculateItemSize:=SizeOf(ResFileItemID) +
                     Length(Name) + 1 +
                     SizeOf(Longint) +
                     SizeOf(Longint) +
                     SizeOf(Boolean) +
                     DataSize;
 end;

function TResFile.Add(const Name: String; const S: PStream): Boolean;
 var
  Item: PResItem;
 begin
  if Name = '' then
   begin
    Add:=False;

    Exit;
   end;

  Item:=Index^.Seek(Name);

  if Item <> nil then
   begin
    Item^.Junk:=True;

    Item^.UpdateHeader(Stream);
   end;

  if Errors^.Count <> 0 then
   Item:=nil
  else
   Item:=Index^.SeekJunk(CalculateItemSize(Name, S^.GetSize));

  if Item = nil then
   begin
    Item:=New(PResItem, Init(Stream^.GetSize, NewStr(Name)));

    Item^.DataSize:=S^.GetSize;
    Item^.Size:=CalculateItemSize(Name, Item^.DataSize);
    Item^.Junk:=False;

    Item^.CalculateDataOffset;

    Item^.UpdateHeader(Stream);

    Index^.Insert(Item);
   end
  else
   begin
    DisposeStr(Item^.Name);

    Item^.Name:=NewStr(Name);

    Item^.DataSize:=S^.GetSize;
    Item^.Junk:=False;

    Item^.CalculateDataOffset;

    Item^.UpdateHeader(Stream);
   end;

  S^.Seek(0);

  Stream^.CopyFrom(S^, Item^.DataSize);

  if Stream^.Status = stOk then
   Add:=True
  else
   begin
    AddError('unable to add "' + Name + '", ds=' + HexL(S^.GetSize) + ', written at ' + HexL(Item^.HeaderOffset));

    Stream^.Reset;

    Add:=False;
   end;

  Flush;
 end;

function TResFile.Remove(const Name: String): Boolean;
 var
  Item: PResItem;
 begin
  Item:=Index^.Seek(Name);

  if Item = nil then
   Remove:=False
  else
   begin
    Item^.DataSize:=0;

    Item^.Junk:=True;

    Item^.UpdateHeader(Stream);

    if Stream^.Status <> stOk then
     begin
      AddError('unable to remove "' + Name + '", stream error ' + HexL(Stream^.Status) + ' at chunk ' +
       HexL(Item^.HeaderOffset));

      Stream^.Reset;

      Remove:=False;
     end
    else
     Remove:=True;

    Flush;

    Index^.Free(Item);
   end;
 end;

function TResFile.Get(const Name: String; const S: PStream): Boolean;
 var
  Item: PResItem;
 begin
  Item:=Index^.Seek(Name);

  if Item = nil then
   Get:=False
  else
   begin
    Stream^.Seek(Item^.DataOffset);

    S^.CopyFrom(Stream^, Item^.DataSize);

    Get:=Stream^.Status = stOk;

    if Stream^.Status <> stOk then
     AddError('failed to get "' + Name + '" at offset ' + HexL(Item^.DataOffset) + ' size ' + HexL(Item^.DataSize));
   end;
 end;

function TResFile.Pack(const TempStream: PStream): Boolean;
 var
  NewIndex: PResIndex;
 procedure DoPack(const Item: PResItem); far;
  var
   SavedSize, SavedHeaderOffset, SavedDataOffset: Longint;
   NewItem: PResItem;
  begin
   if not Item^.Junk then
    begin
     SavedSize:=Item^.Size;
     SavedHeaderOffset:=Item^.HeaderOffset;
     SavedDataOffset:=Item^.DataOffset;

     Item^.Size:=Item^.DataSize + CalculateItemSize(Item^.Name^, 0);
     Item^.HeaderOffset:=TempStream^.GetPos;
     Item^.CalculateDataOffset;

     Item^.UpdateHeaderIndirect(TempStream);

     Stream^.Seek(SavedDataOffset);

     TempStream^.CopyFrom(Stream^, Item^.DataSize);

     NewItem:=New(PResItem, Init(Item^.HeaderOffset, NewStr(GetPString(Item^.Name))));

     NewItem^.DataOffset:=Item^.DataOffset;
     NewItem^.DataSize:=Item^.DataSize;
     NewItem^.Size:=Item^.Size;

     NewIndex^.Insert(NewItem);

     Item^.DataOffset:=SavedDataOffset;
     Item^.HeaderOffset:=SavedHeaderOffset;
     Item^.Size:=SavedSize;
    end;
  end;
 begin
  Pack:=False;

  if Error then
   Exit;

  TempStream^.Seek(0);

  NewIndex:=New(PResIndex, Init($10, $10));

  Index^.ForEach(@DoPack);

  TempStream^.Truncate;

  TempStream^.Flush;

  Stream^.Seek(0);

  TempStream^.Seek(0);

  Stream^.CopyFrom(TempStream^, TempStream^.GetSize);

  Stream^.Truncate;

  Stream^.Flush;

  TempStream^.Seek(0);

  TempStream^.Truncate;

  Pack:=False;

  if Stream^.Status <> stOk then
   AddError('failed to pack, stream error ' + Long2Str(Stream^.Status))
  else
   if TempStream^.Status <> stOk then
    AddError('failed to pack, temp stream error ' + Long2Str(TempStream^.Status))
   else
    begin
     Pack:=True;

     Dispose(Index, Done);

     Index:=NewIndex;

     Exit;
    end;

  Stream^.Reset;

  TempStream^.Reset;
 end;

function TResFile.AddExternal(const Name: String; const S: Pointer; const Linker: PStreamLinker): Boolean;
 var
  Thunk: PStreamThunk;
 begin
  Thunk:=New(PStreamThunk, Init(S, Linker, False));

  AddExternal:=Add(Name, Thunk);

  Dispose(Thunk, Done);
 end;

function TResFile.GetExternal(const Name: String; const S: Pointer; const Linker: PStreamLinker): Boolean;
 var
  Thunk: PStreamThunk;
 begin
  Thunk:=New(PStreamThunk, Init(S, Linker, False));

  GetExternal:=Get(Name, Thunk);

  Dispose(Thunk, Done);
 end;

function TResFile.PackExternal(const S: Pointer; const Linker: PStreamLinker): Boolean;
 var
  Thunk: PStreamThunk;
 begin
  Thunk:=New(PStreamThunk, Init(S, Linker, False));

  PackExternal:=Pack(Thunk);

  Dispose(Thunk, Done);
 end;

function TResFile.Count: Longint;
 begin
  Count:=Index^.Count;
 end;

function TResFile.At(const I: Longint; var Junk: Boolean; var Size: Longint): PString;
 var
  Item: PResItem;
 begin
  Item:=Index^.At(I);

  if Item = nil then
   begin
    Junk:=False;
    Size:=0;
    At:=nil;
   end
  else
   begin
    Junk:=Item^.Junk;
    Size:=Item^.DataSize;
    At:=Item^.Name;
   end;
 end;

{ TResFiles }

function TResFiles.Compare(Key1, Key2: Pointer): xInteger;
 var
  S1, S2: String;
 begin
  S1:=FExpand(PString(Key1)^);
  S2:=FExpand(PString(Key2)^);

  StUpcaseEx(S1);
  StUpcaseEx(S2);

  if S1 < S2 then Compare:=-1 else
  if S1 > S2 then Compare:=1 else
   Compare:=0;
 end;

function TResFiles.KeyOf(Item: Pointer): Pointer;
 begin
  KeyOf:=PResFile(Item)^.FName;
 end;

{ rfInit }

procedure rfInit;
 begin
  ResFiles:=New(PResFiles, Init($10, $10));
 end;

{ rfDone }

procedure rfDone;
 begin
  Dispose(ResFiles, Done);
 end;

end.