unit ScanScan;

interface
uses
     Objects,

     PlugAPI,
     PlugSIDE,
     PlugUTIL,

     jtMHLcmn,

     ScanAPI,
     ScanCmn,

     mbAPI,

     Wizard;

const
 CriticalErrorsCount            = 50;

type
 TScannerStatus = (ssOpening, ssScanning, ssClosing);

 PScanner = ^TScanner;
 TScanner = object(TObject)
  Area: Pointer;
  AreaName: String;
  Full, KillSent, Lock, KillInExportList, AddToImportList: Boolean;
  mb: Pointer;
  No, TotalNo, Current, Total, Scanned, Exported, Killed, Errors: Longint;
  ReportMode: TReportMode;
  mbi: PMessageBaseInterface;

  constructor Init(const AArea: Pointer; const AFull, AKillSent, ALock, AKillInExportList, AAddToImportList: Boolean;
   const AReportMode: TReportMode; const ACurrent, ATotal: Longint);
  procedure Scan;
  procedure Status(const ScannerStatus: TScannerStatus);
  destructor Done; virtual;
 public
  procedure ProcessMessage(const msg: Pointer; var Skip: Boolean);
  procedure ExportMessage(const Amsg: Pointer; var Did: Boolean);
  function CheckAttributes(const Attributes: Longint): Boolean; virtual;
  procedure SetAttributes(var Attributes: Longint); virtual;
  procedure LoadNo; virtual;
  procedure StoreNo; virtual;
 end;

var
 Roller: TRoller;

implementation

{ TScanner }

constructor TScanner.Init(const AArea: Pointer; const AFull, AKillSent, ALock, AKillInExportList, AAddToImportList: Boolean;
  const AReportMode: TReportMode; const ACurrent, ATotal: Longint);
 var
  rc: String;
 begin
  inherited Init;

  Area:=AArea;

  Full:=AFull;
  KillSent:=AKillSent;
  KillInExportList:=AKillInExportList;
  AddToImportList:=AAddToImportList;

  Lock:=ALock;

  Current:=ACurrent;
  Total:=ATotal;

  ReportMode:=AReportMode;

  Scanned:=0;
  Exported:=0;
  Killed:=0;

  GetPStringEx(jt.area.GetName(Area), AreaName);

  mbi:=jt.area.mbi(Area);

  Status(ssOpening);

  if not jt.area.Lock(Area, mb, rc) then
   begin
    jt.LogWrite('?', 'Unable to open ' + AreaName + ': ' + rc);

    Fail;
   end;

  if Lock and not jt.mb.Lock(mb) then
   begin
    jt.LogWrite('?', 'Unable to lock ' + AreaName + ' for scanning');

    jt.area.Unlock(Area, mb);

    Fail;
   end;
 end;

procedure TScanner.Scan;
 const
  Timeout = 100;
 var
  Timer: Longint;
  Skip: Boolean;
  rc: String;
  msg: Pointer;
 begin
  Status(ssScanning);

  LoadNo;

  if No < 0 then
   No:=0;

  Timer:=Clock;

  TotalNo:=jt.mb.Count(mb);

  while (No < TotalNo) and (Errors < CriticalErrorsCount) do
   begin
    if Clock - Timer > Timeout then
     begin
      Status(ssScanning);

      Timer:=Clock;

      if jt.Cancelled then
       Break;
     end;

    if jt.mb.Load(mb, No, msg, rc, mfNone) then
     begin
      Inc(Scanned);

      ProcessMessage(msg, Skip);

      jt.msg.Destroy(msg);

      if not Skip then
       Continue;
     end
    else
     begin
      jt.LogWrite('?', 'Failed to open msg ' + Long2Str(No) + ' in ' + AreaName + ': ' + rc);

      Inc(Errors);
     end;

    Inc(No);

    TotalNo:=jt.mb.Count(mb);
   end;

  Dec(No);

  if No < 0 then
   No:=0;

  StoreNo;

  if Errors >= CriticalErrorsCount then
   jt.Report(AreaName + ': too many serial errors encountered while scanning', True);

  if (ReportMode <> rmNone) and (Exported > 0) then
   begin
    if ReportMode = rmDetail then
     rc:=Long2StrFmt(Scanned) + ' msg' + _s(Scanned) + ' scanned, '
    else
     rc:='';

    rc:=rc + Long2StrFmt(Exported) + ' msg' + _s(Exported) + ' exported';

    if Killed > 0 then
     rc:=rc + ', ' + Long2StrFmt(Killed) + ' msg' + _s(Killed) + ' killed';

    jt.LogWrite('S', AreaName + ': ' + rc);
   end;
 end;

procedure TScanner.Status(const ScannerStatus: TScannerStatus);
 var
  Percent: String;
 begin
  Percent:=jt.Percent(Current, Total);

  case ScannerStatus of
   ssOpening: jt.SetTitle(AreaName + ' ' + Percent + ' (opening)');
   ssScanning: jt.SetTitle(AreaName + ' ' + Percent + ' (scanning)');
   ssClosing: jt.SetTitle(AreaName + ' ' + Percent + ' (closing)');
  end;

  if ScannerStatus = ssScanning then
   jt.SetStatus(AreaName + ' ' + Percent + ' ' + Roller.GetRoller + ' scanning ' + jt.Percent(No, TotalNo));
 end;

destructor TScanner.Done;
 begin
  Status(ssClosing);

  if Lock then
   jt.mb.Unlock(mb);

  jt.area.Unlock(Area, mb);

  inherited Done;
 end;

{ public methods }

procedure TScanner.ProcessMessage(const msg: Pointer; var Skip: Boolean);
 var
  Attributes: Longint;
  Corpseable, Did: Boolean;
  rc: String;
 begin
  Attributes:=jt.msg.GetAttributes(msg);

  if (Attributes and (maSent or maLocal) = maLocal) and CheckAttributes(Attributes) then
   begin
    if not Lock and not jt.mb.Lock(mb) then
     begin
      jt.LogWrite('?', 'Unable to lock ' + AreaName + ' for scanning at msg ' + Long2Str(No));

      Inc(Errors);

      Skip:=True;

      Exit;
     end;

    Corpseable:=Attributes and maKill <> 0;

    ExportMessage(msg, Did);

    if Did then
     begin
      Attributes:=Attributes or maSent;

      SetAttributes(Attributes);

      jt.msg.SetAttributes(msg, Attributes);

      if ReportMode = rmDetail then
       jt.LogWrite('S', AreaName + ': msg #' + Long2Str(jt.mb.NoToGlobal(mb, No)) +
                                   ' from "' + GetPString(jt.msg.GetFromName(msg)) +
                                    '" to "' + GetPString(jt.msg.GetToName(msg)) + '"');

      Inc(Exported);

      if KillSent and Corpseable then
       if jt.mb.Kill(mb, No, rc) then
        begin
         Skip:=False;

         Inc(Killed);

         Errors:=0;
        end
       else
        begin
         jt.LogWrite('?', 'Unable to kill msg ' + Long2Str(No) + ' in ' + AreaName + ': ' + rc);

         Inc(Errors);

         Skip:=True;

         if not jt.mb.Store(mb, No, msg, rc, mfNone) then
          begin
           jt.LogWrite('?', 'Unable to write msg ' + Long2Str(No) + ' in ' + AreaName + ': ' + rc);

           Inc(Errors);
          end;
        end
      else
       begin
        Skip:=True;

        if not jt.mb.Store(mb, No, msg, rc, mfNone) then
         begin
          jt.LogWrite('?', 'Unable to write msg ' + Long2Str(No) + ' in ' + AreaName + ': ' + rc);

          Inc(Errors);
         end
        else
         Errors:=0;
       end;

      if KillInExportList then
       jt.list.Remove(jt.list.ExportList, AreaName);

      if AddToImportList then
       jt.list.Add(jt.list.ImportList, AreaName);
     end
    else
     Skip:=True;

    if not Lock then
     jt.mb.Unlock(mb);
   end
  else
   Skip:=True;
 end;

procedure TScanner.ExportMessage(const Amsg: Pointer; var Did: Boolean);
 var
  msg, msgp: Pointer;
  FromAddress: TAddress;
  Block: TScanMessageBlock;
 begin
  msg:=jt.msg.Clone(Amsg, mcfCloneAll);

  { // area }

  jt.msg.SetArea(msg, AreaName);

  { // attributes }

  jt.msg.SetAttributes(msg, jt.msg.GetAttributes(msg) and not (maSent or maScanned or maLocal));

  { // general hook }

  Block.Area:=Area;
  Block.msg:=msg;
  Block.Did:=False;

  Callbacks[saMessageScanGeneral]^.ChainedPointer(scMessageScanGeneral, @Block, False);

  if Block.Area <> nil then
   begin
    { // tid & fromaddress }

    msgp:=jt.msg.CreateParser(msg, False);

    jt.msg.p.SetKludge(msgp, #1'TID:', #1'TID: ' + jt.core.GetTID);

    ClearAddress(FromAddress);

    jt.msg.p.GetFromAddress(msgp, FromAddress);

    jt.msg.DestroyParser(msgp);

    if IsCleanAddress(FromAddress) then
     begin
      jt.Report('Failed to scan msg ' + Long2Str(No) + ' in ' + AreaName + ': unable to get valid fromaddress', True);

      Inc(Errors);

      Did:=False;
     end
    else
     begin
      jt.msg.ti.PickupTechInfo(msg);

      jt.msg.ti.DoTechInfo(nil, Area, msg, @FromAddress);

      if (Callbacks[saMessageScanFirst]^.ChainedPointer(scMessageScanFirst, @Block, False) = nil) and
         (Callbacks[saMessageScan]^.ChainedPointer(scMessageScan, @Block, False) = nil) and
         (Callbacks[saMessageScanLast]^.ChainedPointer(scMessageScanLast, @Block, False) = nil) then
       begin
        jt.msg.ti.DropTechInfo(msg);

        jt.tossscan.TossMessageToLinks(msg, Area, nil);

        Did:=True;
       end
      else
       Did:=Block.Did;
     end;
   end
  else
   Did:=Block.Did;

  jt.msg.Destroy(msg);
 end;

function TScanner.CheckAttributes(const Attributes: Longint): Boolean;
 begin
  CheckAttributes:=not jt.mb.ExtendedOpened(mb, eooNotScannableAttributes, (@Attributes)^);
 end;

procedure TScanner.SetAttributes(var Attributes: Longint);
 begin
  jt.mb.ExtendedOpened(mb, eooSetScannedAttributes, Attributes);
 end;

procedure TScanner.LoadNo;
 begin
  if Full then
   No:=0
  else
   if jt.mb.ExtendedOpened(mb, eooGetHighWater, No) then
    begin
     HighwaterDebugger.Log('absolute highwater for ' + AreaName + ': ' + Long2Str(No));

     No:=jt.mb.GlobalToNo(mb, No);

     if No = -1 then
      No:=jt.mb.Count(mb) - 1;

     if No > 0 then
      Dec(No);

     HighwaterDebugger.Log('fixed highwater for ' + AreaName + ': ' + Long2Str(No));
    end
   else
    begin
     No:=Longint(jt.st.Get(jt.area.st(Area), 'LastScanned'));

     if No > 0 then
      Dec(No);

     if No <> 0 then
      No:=jt.mb.GlobalToNo(mb, No);
    end;
 end;

procedure TScanner.StoreNo;
 var
  st: Pointer;
  Global: Longint;
 begin
  Global:=jt.mb.NoToGlobal(mb, No);

  if Global < 0 then
   Global:=-1;

  st:=jt.area.st(Area);

  if jt.mb.ExtendedOpened(mb, eooSetHighWater, Global) then
   begin
    HighwaterDebugger.Log('storing global highwater ' + Long2Str(Global) + ' (local ' + Long2Str(No) + ') for ' + AreaName);

    if LongBool(jt.st.Get(st, 'LastScanned')) then
     begin
      jt.st.Remove(st, 'LastScanned');

      jt.area.SetChanged(acfChanged, 'ls removed');
     end;
   end
  else
   if Global <> Longint(jt.st.Get(st, 'LastScanned')) then
    begin
     jt.st.Add(st, 'LastScanned', Pointer(Global), nil);

     jt.area.SetChanged(acfChanged, 'ls changed');
    end;
 end;

end.
